(in-package :earley-parser)

;;;; Object representations -
;;;;-------------------------
(defstruct state
  ;; The "head" of the state
  (condition '? :type string)
  ;; A tree (= list) representing an allowed sequence of successors for the
  ;; condition.
  (subtree)
  ;; Indicates where the dot should be relative to this subtree
  ;; (whats is _observed_ vs what is _exptected_).
  (dot 0)
  ;; Index in sentence where the first in the sequence of
  ;; allowed successors should be.
  (constituent-index 0)
  ;; Index relative to the sentence for where the dot should be.
  (dot-index 0)
  ;; Find out which previous state that led to this state 
  ;; (used for backtracing when creating a tree).
  (source-states))

(defmethod print-object ((state state)(s stream))
  (let ((condition (state-condition state))
        (subtree (state-subtree state))
        (dot (state-dot state)))
    (format s "#{~A~T->~T~{~A ~} . ~{~A ~}~t[~A,~A]}"
            condition (subseq subtree 0 dot) 
	    (subseq subtree dot (length subtree))
            (state-constituent-index state)(state-dot-index state))))

(defstruct chart
  (states))

(defmethod enqueue ((state state) (chart chart))
  (if (member state (chart-states chart) :test *string-comparer*)
      (when (> *debug* 3)
	(format t "  the state ~A is already in the chart~%" state))
      (setf (chart-states chart) (append (chart-states chart) (list state)))))

(defmethod print-object ((chart chart)(s stream))
  (format s "#CHART:")
  (loop for state in (chart-states chart)
     do (format s "~&    ~A~%" state)))

(defstruct chart-listing
  (charts))

(defmethod add-chart ((chart chart) (chart-listing chart-listing))
  (push chart (chart-listing-charts chart-listing)))

(defmethod print-object ((chart-listing chart-listing)(s stream))
  (format s "#CHART-LISTING:")
  (loop for chart in (chart-listing-charts chart-listing)
     and index from 0
     do (format s "~&  ~A. ~A" index chart)))

(defun incomplete? (state)
  "Returns whether or not there is anything left of the subtree behind the dot."
  (not (= (state-dot state) (length (state-subtree state)))))

(defun next-cat (state)
  "Returns the next cathegory of 'state'"
  (let ((subtree (state-subtree state))
        (dot (state-dot state)))
    (when (> (length subtree) dot)
      (nth dot subtree))))


;;;; Earley parser functions -
;;;;--------------------------
(defun predictor (state chart-listing)
  "Predict possible successor states based on the grammar. As a side-effect, add
 these states to the chart that this state belong to."
  (let ((B (next-cat state)) ; The next cathegory for this state.
	;; The index of the first constituent following the current condition.
        (j (state-dot-index state)))
    (loop for production in (gethash B *cfg*)
       collect (let ((new-state
		      (make-state :condition B :subtree production :dot 0 
				  :constituent-index j :dot-index j)))
		 (when (> *debug* 2)
		   (format
		    t "  predictor attempting to enqueue ~A into chart ~A~%" 
		    new-state j))
		 (enqueue 
		  new-state (nth j (chart-listing-charts chart-listing)))
		 new-state))))

(defun scanner (state words chart-listing)
  "Check if the next category for this state is a member of the pos-cathegories
 for the current word. As a side effect, enqueue a new state corresponding to th
is find, into the current chart."
  (let* ((B (next-cat state))
         (j (state-dot-index state))
         (word (nth j words)))
    (when (> *debug* 2)
      (format 
       t "  scanner is considering if ~A is a member of the word-class list for \"~A\" (= ~A)~%"
       B word (gethash word *lexicon*)))
    (when (member B (gethash word *lexicon*) :test #'(lambda (b terminal)
						       (funcall
*string-comparer*
 b
							       (terminal-class terminal))))
      (let ((new-state (make-state :condition B :subtree (list word) :dot 1 
				   :constituent-index j :dot-index (+ j 1))))
        (when (> *debug* 2)
          (format
	   t "  scanner attempting to enqueue ~A into chart ~A~%" 
	   new-state (+ j 1)))
        (enqueue
	 new-state (nth (+ j 1) (chart-listing-charts chart-listing)))))))

(defun completer (state chart-listing)
  "Find and return a list of the previous states that expect this states cathego
ry at this dot-index with the dot moved one step forward. As a side-effect, enqu
eue the states in the current chart."
  (let ((B (state-condition state))
        (j (state-constituent-index state))
        (k (state-dot-index state)))
    (loop for prev-state in (chart-states (nth j (chart-listing-charts
						  chart-listing)))
       when (when t
	      (let ((A (next-cat prev-state))) 
		(when (funcall *string-comparer* A B)
		  (when (> *debug* 3)
		    (format
		     t "    completer found the state: ~A to match ~A~%" 
		     prev-state state))
		  t)))
       collect (let ((new-state (make-state
				 :condition (state-condition prev-state)
				 :subtree (state-subtree prev-state)
				 :dot (+ (state-dot prev-state) 1)
				 :constituent-index (state-constituent-index
						     prev-state)
				 :dot-index k
				 :source-states (append (list state)
							(state-source-states
							 prev-state)))))
		 (when (> *debug* 2)
		   (format 
		    t "  completer attempting to enqueue ~A into chart ~A~%" 
		    new-state k))
		 (enqueue 
		  new-state (nth k (chart-listing-charts chart-listing)))
		 new-state))))

(defun earley-parse (sentence &optional (split-regex " "))
  "Convert a string of words into a chart conforming to the grammar."
  (let ((words (remove "" (cl-ppcre:split split-regex sentence)))
        (chart-listing (make-chart-listing)))
    ;; Initialize charts, one chart per word in the sentence
    (loop for i from 0 to (length words)
       do (add-chart (make-chart) chart-listing))
    ;; Start off by enqueuing a dummy state in the first chart
    (enqueue (make-state :condition "G" :subtree '("S") :dot-index 0)
             (nth 0 (chart-listing-charts chart-listing)))
    ;; Then for each chart (= one per word)...
    (loop for chart in (chart-listing-charts chart-listing) and
       chart-index from 0
       ;; And for each possible state in that chart
       do (progn
	    (when (> *debug* 0)
	      (format t "~%---- processing chart #~A ----~%" chart-index))
	    (loop for state-index from 0
	       until (>= state-index (length (chart-states chart)))
	       do (let ((state (nth state-index (chart-states chart))))
		    (when (> *debug* 1)
		      (format t "considering state: ~A~%" state)
		      (format t "next cat of this~Astate is ~A~%"
			      (if (incomplete? state) " (incomplete) " " ")
			      (next-cat state)))
		    (cond ((and (incomplete? state) 
				(not (member (next-cat state)
					     *part-of-speech* :test *string-comparer*)))
			   (when (> *debug* 1)
			     (format t "predicting...~%"))
			   (predictor state chart-listing))
			  ((and (incomplete? state) 
				(member (next-cat state) *part-of-speech*
					:test *string-comparer*))
			   (unless (eq chart (first
					      (last
					       (chart-listing-charts
						chart-listing))))
			     (when (> *debug* 1)
			       (format t "scanning...~%"))
			     (scanner state words chart-listing)))
			  (t
			   (when (> *debug* 1)
			     (format t "completing...~%"))
			   (completer state chart-listing)))))
	    (when (> *debug* 0)
	      (format t "~%"))))
    chart-listing))


;;;; Tree construction functions
;;;;----------------------------
(defun state->tree (state)
  "Creates a tree from a chart-listing object containting charts"
  (if (null (state-source-states state))
    (list (state-condition state) (first (state-subtree state)))
    (cons (state-condition state)
          (reverse (loop for state in (state-source-states state)
                         collect (state->tree state))))))

(defun chart-listing->trees (chart-listing)
  "Return a list of trees created by following each successful parse in the last
 chart of 'chart-listings'"
  (loop for state in (chart-states 
		      (first (last (chart-listing-charts chart-listing))))
     when (and (funcall *string-comparer* (state-condition state) "S")
	       (= (state-constituent-index state) 0)
	       (= (state-dot-index state)
		  (- (length (chart-listing-charts chart-listing)) 1))
	       (not (incomplete? state)))
     collect (state->tree state)))
